# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
Main file that parses the arguments and sets up the REST handler.
'''

import argparse
import sys
from workload_auto import dcnm_top
from workload_auto import logger
from workload_auto import utils

class WlAuto:
    '''
    Main worlkoad automation class.
    '''
    def __init__(self):
        '''
        Init routine to parse the arguments and calls the top level Class
        apart from setting up the REST handler.
        '''
        parser = argparse.ArgumentParser()
        parser.add_argument("-c", "--config", dest="config",
                            default='/etc/vmm_workload_auto/conf.yml',
                            help="Provide configuration file", metavar="FILE")
        args = parser.parse_args()
        if args.config is None:
            print("config file not present")
            return
        print("Config file is " + args.config)
        file_dict, exc = utils.yml_file_read(args.config)
        if not file_dict:
            print("Exception in config file read ", exc)
            return
        print("file content is ", file_dict)
        log_file = file_dict.get('LogFile', '/tmp/tmplog')
        logger.set_logging(log_file)
        dcnm_top.DcnmTop(file_dict)

def wl_auto_main():
    '''
    Main init routine.
    '''
    WlAuto()

if __name__ == '__main__':
    sys.exit(WlAuto())



# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
File containing REST handlers.
'''

from flask import Flask, json

companies = [{"id": 1, "name": "Company One"}, {"id": 2, "name": "Company Two"}]

api = Flask(__name__)

@api.route('/workload_auto/clean', methods=['POST'])
def workload_clean():
    '''
    Handler to clean up the network attachments.
    '''
    GLOB_CB("CLEAN", None)
    return json.dumps({"success": True}), 201

@api.route('/workload_auto/refresh', methods=['POST'])
def workload_refresh():
    '''
    Handler to re-read the config file contents and re-apply the network
    attachments.
    '''
    GLOB_CB("REFRESH", None)
    return json.dumps({"success": True}), 201

@api.route('/workload_auto/resync', methods=['POST'])
def workload_resync():
    '''
    Handler to re-discover the neighbours and redo the network attachments
    if needed.
    '''
    GLOB_CB("RESYNC", None)
    return json.dumps({"success": True}), 201

GLOB_CB = None

def rest_init(port, cb_arg):
    '''
    Top level REST init routine.
    '''
    global GLOB_CB

    GLOB_CB = cb_arg
    api.run(host='0.0.0.0', port=port)



# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
Vsphere module, which is instantiated one per vCenter per DCNM. This maintains
the list of DVS and PG's and the associated, pNIC's, VLAN's and neighbour for
each pNIC. This module also calls the DCNM helper to do the network
attachment.
'''

from functools import wraps
import sys
from workload_auto import logger
from workload_auto import vsphere_helper

LOG = logger.get_logging(__name__)

class HostInfo:
    '''
    Class to store the host's neighbour information.
    '''
    def __init__(self, host, sw, dcnm_obj):
        '''
        Initialization routine for the class.
        '''
        self.host = host
        self.dcnm_obj = dcnm_obj
        self.neighbour_list = []
        self.add_neighbour(sw)

    def fill_po_intf(self):
        '''
        This function finds the associated PO index associated with the
        interfaces, if any.
        '''
        ind = 0
        for nbr_dict in self.neighbour_list:
            snum = nbr_dict.get('snum')
            intf_list = nbr_dict.get('sw_port')
            new_intf_list = []
            for intf in intf_list.split(","):
                po_id = self.dcnm_obj.get_portchannel_id(snum, intf)
                if po_id is None:
                    new_intf_list.append(intf)
                else:
                    new_intf = "Port-channel" + str(po_id)
                    if new_intf not in new_intf_list:
                        new_intf_list.append(new_intf)
            new_intf_str = ",".join(new_intf_list)
            nbr_dict['sw_port'] = new_intf_str
            self.neighbour_list[ind] = nbr_dict
            ind = ind + 1

    def add_neighbour(self, sw_info):
        '''
        This function maintains the list of hosts and its neighbours.
        '''
        flag = False
        cnt = 0
        for nbr_dict in self.neighbour_list:
            snum = nbr_dict.get('snum')
            if sw_info.get('snum') == snum:
                flag = True
                break
            cnt = cnt + 1
        nei_dict = sw_info
        if flag:
            LOG.info("Overwritting the neighbour for %s", sw_info.get('snum'))
            self.neighbour_list[cnt] = nei_dict
        else:
            self.neighbour_list.append(nei_dict)
        self.fill_po_intf()

    def add_undiscovered_neighbour(self, snum):
        '''
        This function adds an undiscovered vpc host to the neighbours list.
        Except for snum, other information are just copied over.
        '''
        if len(self.neighbour_list) == 0:
            LOG.error("Cannot add an undiscovered neighbour to an empty list")
            return
        nbr_dict = self.neighbour_list[0]
        new_dict = nbr_dict.copy()
        new_dict['snum'] = snum
        self.neighbour_list.append(new_dict)

    def get_neighbour(self):
        '''
        Function to return the neighbour list
        '''
        return self.neighbour_list

class VsphHandler:
    '''
    vSphere handler class.
    '''
    def __init__(self, **kwargs):
        '''
        The init routine which also instantiates the vsphere helper class.
        The two main data structures are:
        dvs_dict => Map of DVS pointing to map of DVS PG's.
        host_dict => Map of host name pointing to map of host PG's.
        Each DVS PG or host PG holds a list of fabric structure.
        Each fabric structure has the mapped DCNM Network name, VLAN and a
        map of hosts that this DVS spans.
        Each host map contains the object for the class HostInfo which holds
        the list of neighbour switches. It's a list because one host can point
        to multiple switches (vPC).
        '''
        srvr_cntrl_dict = kwargs['vsph_dict']
        self.ip_addr = srvr_cntrl_dict['Ip']
        self.user = srvr_cntrl_dict['User']
        self.pwd = srvr_cntrl_dict['Password']
        self.dcnm_obj = kwargs['dcnm_obj']
        self.csv_list = {}
        self.fab_parent_info = {}
        self.dvs_dict = {}
        self.host_dict = {}
        self.vsph_obj = vsphere_helper.VsphereHelper(ip=self.ip_addr,
                                                     user=self.user,
                                                     pwd=self.pwd)

    def get_ip(self):
        '''
        Retrieve the IP address of vSphere
        '''
        return self.ip_addr

    def get_dvs(self, dvs_name):
        '''
        Returns the dict of DVS PG's when given the DVS name.
        '''
        return self.dvs_dict.get(dvs_name)

    def set_dvs(self, dvs_name):
        '''
        Initializes the DVS dict if empty.
        '''
        dvs_dict = self.get_dvs(dvs_name)
        if dvs_dict is None:
            self.dvs_dict.update({dvs_name: {}})

    def get_host(self, host_name):
        '''
        Returns the dict of Host PG's when given the DVS name.
        '''
        return self.host_dict.get(host_name)

    def set_host(self, host_name):
        '''
        Initializes the Host dict if empty.
        '''
        host_dict = self.get_host(host_name)
        if host_dict is None:
            self.host_dict.update({host_name: {}})

    def cmn_dict_decor(fnc):
        '''
        Decorator function that calls the original function passing
        either the dvs_dict or host_dict based on the argument.
        '''
        @wraps(fnc)
        def cmn_dict_decor_int(self, *args):
            is_dvs = args[0]
            if is_dvs:
                cmn_dict = self.get_dvs(args[1])
            else:
                cmn_dict = self.get_host(args[1])
            return fnc(self, *args, cmn_dict=cmn_dict)
        return cmn_dict_decor_int

    @cmn_dict_decor
    def get_cmn_pg(self, is_dvs, dvs_or_host_name, cmn_pg_name, cmn_dict=None):
        '''
        Common function to return the port group or DVS PG content (List of fab)
        '''
        if cmn_dict is None:
            return None
        return cmn_dict.get(cmn_pg_name)

    @cmn_dict_decor
    def set_cmn_pg(self, is_dvs, dvs_or_host_name, cmn_pg_name, cmn_dict=None):
        '''
        Common function to initialzie the port group or DVS PG map.
        '''
        if cmn_dict is None:
            LOG.error("Unable to set dvs %(dvs)s dvs_name is null",
                      {'dvs': dvs_or_host_name})
            return
        cmn_pg = self.get_cmn_pg(is_dvs, dvs_or_host_name, cmn_pg_name)
        if cmn_pg is None:
            cmn_dict.update({cmn_pg_name: []})

    @cmn_dict_decor
    def get_cmn_fabric_info(self, is_dvs, dvs_or_host_name, cmn_pg_name, fab,
                            cmn_dict=None):
        '''
        Common function to return the fabric content of a PG or DVS PG.
        '''
        cmn_pg_list = self.get_cmn_pg(is_dvs, dvs_or_host_name, cmn_pg_name)
        if cmn_pg_list is None:
            return None
        #dvs_pg_list is basically a list of fabric dict
        for cmn_fab_dict in cmn_pg_list:
            if cmn_fab_dict.get('fabric') == fab:
                return cmn_fab_dict
        return None

    @cmn_dict_decor
    def set_cmn_fabric_info(self, is_dvs, dvs_or_host_name, cmn_pg_name, fab,
                            network, cmn_dict=None):
        '''
        Common function to set the fabric content of a PG or DVS PG.
        '''
        fabric_dict = self.get_cmn_fabric_info(is_dvs, dvs_or_host_name,
                                               cmn_pg_name, fab)
        if fabric_dict is not None:
            LOG.error('Fabric info already present for %(fabric)s, duplicate!!',
                      {'fabric': fab})
            return
        fabric_dict = {}
        # This is set at the beginning as well as when neighbour is filled.
        vlan = self.vsph_obj.get_vlan(is_dvs, dvs_or_host_name, cmn_pg_name)
        fabric_dict.update({'fabric': fab, 'network': network, 'vlan': vlan,
                            'host_obj_map': {}})
        cmn_dict[cmn_pg_name].append(fabric_dict)

    @cmn_dict_decor
    def set_neighbour_host(self, is_dvs, dvs_or_host_name, cmn_pg_name, fab,
                           host, sw_info, cmn_dict=None):
        '''
        Common function to set the host map of the fabric structure inside
        DVS PG or PG. This initializes the HostInfo class for the host.
        '''
        fabric_dict = self.get_cmn_fabric_info(is_dvs, dvs_or_host_name,
                                               cmn_pg_name, fab)
        if fabric_dict is None:
            LOG.error('Fabric info not present for %(fabric)s',
                      {'fabric': fab})
            return
        host_obj_dict = fabric_dict.get('host_obj_map')
        if host_obj_dict is None:
            host_obj_dict = {}
        if host not in host_obj_dict:
            host_obj = HostInfo(host, sw_info, self.dcnm_obj)
            host_obj_dict.update({host: host_obj})
            fabric_dict.update({'host_obj_map': host_obj_dict})
        else:
            host_obj = host_obj_dict.get(host)
            host_obj.add_neighbour(sw_info)

    @cmn_dict_decor
    def set_undisc_neighbour_host(self, is_dvs, dvs_or_host_name, cmn_pg_name,
                                  fab, host, snum, cmn_dict=None):
        '''
        In some cases, like a ESXI host connecting to a FEX, which is dual homed
        to two vPC leafs, the ESXi will only see one neighbour. But, the config
        needs to happen in both the leafs. This function sets the undiscovered
        vpc host also as one of the neighbour.
        '''
        fabric_dict = self.get_cmn_fabric_info(is_dvs, dvs_or_host_name,
                                               cmn_pg_name, fab)
        if fabric_dict is None:
            LOG.error('Fabric info not present for %(fabric)s',
                      {'fabric': fab})
            return
        host_obj_dict = fabric_dict.get('host_obj_map')
        if host_obj_dict is None or host not in host_obj_dict:
            LOG.error("host %(host)s does not have a neighbour, so cannot set "
                      "an undiscovered host ", host)
        else:
            host_obj = host_obj_dict.get(host)
            host_obj.add_undiscovered_neighbour(snum)

    @cmn_dict_decor
    def set_vlan(self, is_dvs, dvs_or_host_name, cmn_pg_name, fab, vlan,
                 cmn_dict=None):
        '''
        Common function to set the VLAN of the fabric structure inside
        DVS PG or PG.
        '''
        fabric_dict = self.get_cmn_fabric_info(is_dvs, dvs_or_host_name,
                                               cmn_pg_name, fab)
        if fabric_dict is None:
            LOG.error('Fabric info not present for %(fabric)s',
                      {'fabric': fab})
            return
        fabric_dict.update({'vlan': vlan})

    @cmn_dict_decor
    def get_vlan(self, is_dvs, dvs_or_host_name, cmn_pg_name, fab,
                 cmn_dict=None):
        '''
        Common function to vlan the VLAN from the fabric structure inside
        DVS PG or PG.
        '''
        fabric_dict = self.get_cmn_fabric_info(is_dvs, dvs_or_host_name,
                                               cmn_pg_name, fab)
        if fabric_dict is None:
            LOG.error('Fabric info not present for %(fabric)s',
                      {'fabric': fab})
            return 0
        return fabric_dict.get('vlan')

    def add_dvs_or_host(self, is_dvs, dvs_or_host, cmn_pg, fabric, network):
        '''
        Add Top level DVS or Host map.
        '''
        if cmn_pg is None or fabric is None or network is None:
            LOG.error('Invalid CSV values, %(cmn_pg)s %(fabric)s %(network)s',
                      {'cmn_pg': cmn_pg, 'fabric': fabric, 'network': network})
            return
        if is_dvs:
            self.set_dvs(dvs_or_host)
        else:
            self.set_host(dvs_or_host)
        #self.set_dvs_pg(dvs, dvs_pg)
        self.set_cmn_pg(is_dvs, dvs_or_host, cmn_pg)
        self.set_cmn_fabric_info(is_dvs, dvs_or_host, cmn_pg, fabric, network)

    def validate_cfg(self, csv_list):
        '''
        Function to validate if the DVS, DVS PG, Host or Host PG specified in
        CSV map, exists in vCenter.
        '''
        LOG.info("Validating vCenter params in %s", self.get_ip)
        for csv_map in csv_list:
            vcenter_ip = csv_map.get('vCenter')
            if vcenter_ip is None or vcenter_ip != self.get_ip():
                continue
            if csv_map.get('Dvs') is not None and csv_map.get('Dvs') != '':
                if not self.vsph_obj.is_dvs_dvspg_exist(csv_map.get('Dvs'),
                                                        csv_map.get('Dvs_pg')):
                    LOG.error("DVS %s or DVS PG %s does not exist in vCenter",
                              csv_map.get('Dvs'), csv_map.get('Dvs_pg'))
                    return False
            elif csv_map.get('Host') is not None and csv_map.get('Host') != '':
                if not self.vsph_obj.is_host_hostpg_exist(
                        csv_map.get('Host'), csv_map.get('Host_pg')):
                    LOG.error("Host %s or Host PG %s does not exist in vCenter",
                              csv_map.get('Host'), csv_map.get('Host_pg'))
                    return False
        return True

    def store_csv_map(self):
        '''
        Top level function that creates the DVS or Host map based on the
        contents in the spec file.
        '''
        for csv_map in self.csv_list:
            vcenter_ip = csv_map.get('vCenter')
            if vcenter_ip is None or vcenter_ip != self.get_ip():
                continue
            if csv_map.get('Dvs') is not None and csv_map.get('Dvs') != '':
                self.add_dvs_or_host(True, csv_map.get('Dvs'),
                                     csv_map.get('Dvs_pg'),
                                     csv_map.get('Fabric'),
                                     csv_map.get('Network'))
            elif csv_map.get('Host') is not None and csv_map.get('Host') != '':
                self.add_dvs_or_host(False, csv_map.get('Host'),
                                     csv_map.get('Host_pg'),
                                     csv_map.get('Fabric'),
                                     csv_map.get('Network'))
            else:
                LOG.error("Both DVS and Host values are empty in CSV")
        LOG.info("----After storing CSV file----")
        LOG.info("DVS Dict is %s", self.dvs_dict)
        LOG.info("Host Dict is %s", self.host_dict)

    def _is_peer_discovered(self, peer_snum, sw_nbr):
        '''
        Check if the peer VPC is present in the list of neighbours.
        '''
        for _sw in sw_nbr:
            _snum = _sw.get('snum')
            if peer_snum == _snum:
                return True
        return False

    def _fill_host_neighbour_info(self, is_dvs, dvs_or_host_dict):
        '''
        This function fills the neighbour information for each host.
        For every host, it queries the vsphere object to get the list of
        pNics, vlan. For pNic it finds the list of neighbouring switches and
        calls the appropriate set function to populate the values.
        '''
        for dvs_or_host_name, cmn_dict in dvs_or_host_dict.items():
            for cmn_pg, cmn_pg_list in cmn_dict.items():
                vlan = self.vsph_obj.get_vlan(is_dvs, dvs_or_host_name, cmn_pg)
                host_pnic_dict = self.vsph_obj.get_pnic(is_dvs,
                                                        dvs_or_host_name,
                                                        cmn_pg)
                for fab_dict in cmn_pg_list:
                    fab = fab_dict.get('fabric')
                    # This is a dynamic data, means it can change.
                    # So, vlan is also re-discovered and set here.
                    self.set_vlan(is_dvs, dvs_or_host_name, cmn_pg, fab, vlan)
                    sw_inv = self.dcnm_obj.get_switches(fab)
                    for host, sw_nbr in host_pnic_dict.items():
                        _vpc_map = {}
                        undisc_nei = []
                        for _sw in sw_nbr:
                            _snum = _sw.get('snum')
                            if _snum not in sw_inv:
                                LOG.error("Neighbor snum %(snum)s not in "
                                          "inventory, skiping", {'snum': _snum})
                                continue
                            _sw_info = sw_inv.get(_snum)
                            self.set_neighbour_host(
                                is_dvs, dvs_or_host_name, cmn_pg, fab,
                                host, _sw)
                            if not _sw_info.get('is_vpc'):
                                continue
                            #VPC case
                            peer_snum = _sw_info.get('peer_snum')
                            if not self._is_peer_discovered(peer_snum, sw_nbr):
                                LOG.info("snum %(snum)s vpc pair "
                                         "%(vpc)s not recognized as a "
                                         "neighbour, may be a single fex case",
                                         {'snum': _snum, 'vpc': peer_snum})
                                undisc_nei.append(peer_snum)
                        for _nb_snum in undisc_nei:
                            LOG.info("Setting undiscovered vpc neighbour %s",
                                     _nb_snum)
                            self.set_undisc_neighbour_host(
                                is_dvs, dvs_or_host_name, cmn_pg, fab, host,
                                _nb_snum)

    def fill_host_neighbour_info(self, is_dvs):
        '''
        Top level function for filling the neighbour info for either DVS or
        Host.
        '''
        if is_dvs:
            cmn_dict = self.dvs_dict
        else:
            cmn_dict = self.host_dict
        self._fill_host_neighbour_info(is_dvs, cmn_dict)

    def _create_sw_nw_dict(self, vlan, sw_intf_dict):
        '''
        Create a map of snum's containing vlan and interface list.
        '''
        sw_nw_dict = {}
        for snum, intf_list in sw_intf_dict.items():
            new_nw_str = {'vlan': vlan, 'intf_list': intf_list}
            sw_nw_dict.update({snum: new_nw_str})
        return sw_nw_dict

    def _summarize_validate_cfg(self, cfg_fab_dict, vlan, sw_intf_dict, nwk,
                                fab_name):
        '''
        Validates the config that needs to be sent to DCNM. This also
        aggregates the interface list to the existing sw map present in every
        network structure.
        '''
        nwk_exist_dict = {}
        # (fixme)  this loop below is not yet UT'ed.
        if fab_name in cfg_fab_dict:
            nwk_exist_dict = cfg_fab_dict.get(fab_name)
            if nwk in nwk_exist_dict:
                nwk_exist_sw_cfg = nwk_exist_dict.get(nwk)
                for snum, intf_list in sw_intf_dict.items():
                    if snum in nwk_exist_sw_cfg:
                        sw_cfg = nwk_exist_sw_cfg.get(snum)
                        if vlan != sw_cfg.get('vlan'):
                            LOG.error("Conflict VLAN value of %(vlan)s, "
                                      "%(exist_vlan)s for fab %(fab)s "
                                      "nwk %(nwk)s, snum %(snum)s",
                                      {'vlan': vlan,
                                       'exist_vlan': sw_cfg.get('vlan'),
                                       'fab': fab_name, 'nwk': nwk,
                                       'snum': snum})
                            return False, {}
                        new_intf_list = sw_cfg.get('intf_list') + intf_list
                        nwk_exist_sw_cfg[snum].update(
                            {'intf_list': new_intf_list})
                    else:
                        new_nw_str = {'vlan': vlan, 'intf_list': intf_list}
                        nwk_exist_sw_cfg.update({snum: new_nw_str})
                sw_nw_dict = nwk_exist_sw_cfg
            else:
                sw_nw_dict = self._create_sw_nw_dict(vlan, sw_intf_dict)
        else:
            sw_nw_dict = self._create_sw_nw_dict(vlan, sw_intf_dict)
        nwk_exist_dict.update({nwk: sw_nw_dict})
        cfg_fab_dict.update({fab_name: nwk_exist_dict})
        return True, cfg_fab_dict

    def _agg_cfg_data(self, cfg_fab_dict, dvs_or_host_dict):
        '''
        Aggregates the information that is needed according to DCNM format.
        Basically this creates a map of fabric, that has a map of networks
        with each network having a list containing the VLAN and a map of switch
        interface information.
        '''
        for dvs_or_host_name, cmn_dict in dvs_or_host_dict.items():
            for cmn_pg, cmn_pg_list in cmn_dict.items():
                for fab_dict in cmn_pg_list:
                    cfg_dict = {}
                    nwk = fab_dict.get('network')
                    fab_name = fab_dict.get('fabric')
                    vlan = fab_dict.get('vlan')
                    for host, obj in fab_dict.get('host_obj_map').items():
                        nbr_list = obj.get_neighbour()
                        for nbr_dict in nbr_list:
                            snum = nbr_dict.get('snum')
                            if snum not in cfg_dict:
                                cfg_dict.update({snum: []})
                            cfg_dict[snum].append(nbr_dict.get('sw_port'))
                    if len(cfg_dict) == 0:
                        LOG.info("No neighbours for fab %(fab)s, nwk %(nwk)s",
                                 {'fab': fab_name, 'nwk': nwk})
                        continue
                    ret, cfg_fab_dict = self._summarize_validate_cfg(
                        cfg_fab_dict, vlan, cfg_dict, nwk, fab_name)
                    if not ret:
                        LOG.error("Incorrect config ")
                        sys.exit(1)
                    #self.dcnm_obj.attach_network(fab_parent, fab_name, nwk,
                    #                            cfg_dict, vlan, enable,
                    #                            auto_deploy)
        return cfg_fab_dict

    def _create_cfg_data(self, cfg_fab_dict, is_dvs):
        '''
        Top level function to create the config data to be sent to DCNM for
        attach. This creates the config data based on dvs_dict or host_dict.
        '''
        if is_dvs:
            cmn_dict = self.dvs_dict
        else:
            cmn_dict = self.host_dict
        return self._agg_cfg_data(cfg_fab_dict, cmn_dict)

    def _cfg_all(self, enable=True, auto_deploy=False):
        '''
        Function that creates the config data for both DVS and host.
        For every fabric, it calls the DCNM helper to attach the network.
        '''
        cfg_fab_dict = {}
        cfg_fab_dict = self._create_cfg_data(cfg_fab_dict, True)
        cfg_fab_dict = self._create_cfg_data(cfg_fab_dict, False)
        for fab_name, fab_items in cfg_fab_dict.items():
            if fab_name in self.fab_parent_info:
                fab_parent = self.fab_parent_info.get(fab_name)
            else:
                fab_parent = fab_name
            for nwk, nwk_items in fab_items.items():
                #vlan = nwk_items[0]
                cfg_dict = nwk_items
                arg_dict = {'fab_parent': fab_parent, 'fab': fab_name,
                            'nwk': nwk, 'snum_vlan_intf_dict': cfg_dict}
                LOG.info("Calling attach network for enable %(enable)s "
                         "arg %(arg)s", {'enable': enable, 'arg': arg_dict})
                self.dcnm_obj.attach_network(enable, auto_deploy, arg_dict)
                #self.dcnm_obj.attach_network(fab_parent, fab_name, nwk,
                #                            cfg_dict, vlan, enable,
                #                            auto_deploy)

    def configure(self, read_cfg=False, enable=True, auto_deploy=False,
                  csv_list=[]):
        '''
        Top level configure function that does the following:
        refresh => Read cfg, read neighbour and reconfigure
        resync => read neighbour and reconfigure
        cleanup => reconfigure
        '''
        if read_cfg:
            self.csv_list = csv_list
            self.store_csv_map()
        if enable:
            self.fill_host_neighbour_info(True)
            self.fill_host_neighbour_info(False)
        self.fab_parent_info = self.dcnm_obj.get_all_fabric_associations()
        self._cfg_all(enable=enable, auto_deploy=auto_deploy)
        # Resetting it back so that it's not used from the cache.
        self.fab_parent_info = {}



# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# pylint: disable=no-member
'''
DCNM helper module. This module interacts with DCNM.
'''

from functools import wraps
import json
import requests
from requests.exceptions import HTTPError
from workload_auto import logger

LOG = logger.get_logging(__name__)
ADD = "ADD"
DELETE_ADD = "DELETE_ADD"
NOOP = "NOOP"

class DcnmHelper:
    '''
    DCNM helper class.
    '''
    def __init__(self, **kwargs):
        '''
        Init routine that initializes the URL's, user, pws etc.
        '''
        self._ip = kwargs['ip']
        self._user = kwargs['user']
        self._pwd = kwargs['pwd']
        self._timeout_resp = 10
        self._req_headers = {'Accept': 'application/json',
                             'Content-Type': 'application/json; charset=UTF-8'}
        self._resp_ok = (requests.codes.ok, requests.codes.created,
                         requests.codes.accepted)
        self._expiration_time = 100000
        self._protocol_host_url = "https://" + self._ip + "/"
        self._base_url = "rest/top-down/fabrics/"
        self._attach_url = "networks/attachments"
        self._nwk_get_url = "rest/top-down/fabrics/"
        self._deploy_url = "networks/deployments"
        self._inventory_url = "rest/control/fabrics/"
        self._get_fab_association_url = "rest/control/fabrics/msd/fabric-associations/"
        self._get_pc_id_url = "rest/interface/pcid"
        LOG.info("Initialized DCNM helper class")

    def _build_url(self, remaining_url):
        '''
        Appends the base URL with the passing URL.
        '''
        return self._protocol_host_url + remaining_url

    def http_exc_handler(http_func):
        '''
        Decorator function for catching exceptions.
        '''
        @wraps(http_func)
        def exc_handler_int(*args):
            try:
                fn_name = http_func.__name__
                return http_func(*args)
            except HTTPError as http_err:
                LOG.error("HTTP error during call to %(func)s, %(err)s",
                          {'func': fn_name, 'err': http_err})
        return exc_handler_int

    @http_exc_handler
    def login(self):
        '''
        Function for login to DCNM.
        '''
        login_url = self._build_url('rest/logon')
        payload = {'expirationTime': self._expiration_time}
        res = requests.post(login_url, data=json.dumps(payload),
                            headers=self._req_headers,
                            auth=(self._user, self._pwd),
                            timeout=self._timeout_resp, verify=False)
        session_id = ""
        if res and res.status_code in self._resp_ok:
            session_id = res.json().get('Dcnm-Token')
            self._req_headers.update({'Dcnm-Token': session_id})
            return True
        LOG.error("Login failed with status %(status)s",
                  {'status': res.status_code})
        return False

    @http_exc_handler
    def logout(self):
        '''
        Function for logoff from DCNM.
        '''
        logout_url = self._build_url('rest/logout')
        requests.post(logout_url, headers=self._req_headers,
                      timeout=self._timeout_resp, verify=False)

    def _get_snum_from_name(self, name, data):
        '''
        Returns the serial number from the name
        '''
        for sw_info in data:
            if sw_info.get('logicalName') == name:
                return sw_info.get('serialNumber')
        LOG.error("Unable to get snum for switch %(name)s", {'name': name})
        return None

    @http_exc_handler
    def _get_switches(self, fabric):
        '''
        Function for retrieving the switch list from DCNM, given the fabric.
        '''
        sw_map = {}
        url = self._build_url(self._inventory_url) + fabric + "/inventory/"
        res = requests.get(url, headers=self._req_headers,
                           timeout=self._timeout_resp, verify=False)
        if res and res.status_code in self._resp_ok:
            data = res.json()
            for sw_info in data:
                if 'serialNumber' in sw_info:
                    snum = sw_info.get('serialNumber')
                    name = sw_info.get('logicalName')
                    is_vpc = sw_info.get('isVpcConfigured')
                    peer = ""
                    peer_snum = ""
                    if is_vpc is None or not is_vpc:
                        is_vpc = False
                    else:
                        peer = sw_info.get('peer')
                        peer_snum = self._get_snum_from_name(peer, data)
                    vpc_map = {'name': name, 'is_vpc': is_vpc, 'peer': peer,
                               'peer_snum': peer_snum}
                    sw_map.update({snum: vpc_map})
        else:
            LOG.error("invalid result for get_switches status %(status)s",
                      {'status': res.status_code})
        return sw_map

    def get_switches(self, fabric):
        '''
        Top level function for retrieving the switches.
        '''
        sw_info = []
        try:
            ret = self.login()
            if ret:
                sw_info = self._get_switches(fabric)
                self.logout()
        except Exception as exc:
            LOG.error("Exception in get_switches, %(exc)s", {exc:exc})
        return sw_info

    @http_exc_handler
    def _get_portchannel_id(self, snum, intf):
        '''
        Retrieve the PO-ID associated with the interface, if any.
        '''
        url = self._build_url(self._get_pc_id_url) + "?serialNumber=" + (
            snum + "&ifName=" + intf)
        res = requests.get(url, headers=self._req_headers,
                           timeout=self._timeout_resp, verify=False)
        data = None
        if res and res.status_code in self._resp_ok:
            if len(res.content) != 0:
                data = res.json()
        else:
            LOG.error("invalid result for get_pc_id status %(status)s",
                      {'status': res.status_code})
        return data

    def get_portchannel_id(self, snum, intf):
        '''
        Top level function for retrieving the PO id.
        '''
        try:
            ret = self.login()
            if ret:
                pc_id = self._get_portchannel_id(snum, intf)
                self.logout()
        except Exception as exc:
            LOG.error("Exception in get_pc_id, %(exc)s", {exc:exc})
        return pc_id

    @http_exc_handler
    def _get_all_fabric_associations(self):
        '''
        Retrieve the parent fabric in case of MSD's.
        '''
        url = self._build_url(self._get_fab_association_url)
        res = requests.get(url, headers=self._req_headers,
                           timeout=self._timeout_resp, verify=False)
        if res and res.status_code in self._resp_ok:
            data = res.json()
            fab_dict = {}
            for fab_elem in data:
                par = fab_elem.get('fabricParent')
                if par is not None and par != 'None':
                    fab_dict.update({fab_elem.get('fabricName'): fab_elem.get(
                                                'fabricParent')})
        else:
            LOG.error("invalid result for get_fabric_assoc status %(status)s",
                      {'status': res.status_code})
        return fab_dict

    def get_all_fabric_associations(self):
        '''
        Top level function to get the parent fabric in MSD.
        '''
        fab_parent = None
        try:
            ret = self.login()
            if ret:
                fab_parent = self._get_all_fabric_associations()
                self.logout()
        except Exception as exc:
            LOG.error("Exception in get_fabric_associations, %(exc)s",
                      {exc:exc})
        return fab_parent

    @http_exc_handler
    def _get_attachments(self, fabric, nwk_name):
        '''
        Retrieve the network attachment given the fabric and network.
        '''
        url = self._build_url(self._base_url) + fabric + "/" + (
            self._attach_url + "?network-names=" + nwk_name)
        res = requests.get(url, headers=self._req_headers,
                           timeout=self._timeout_resp, verify=False)
        if res and res.status_code in self._resp_ok:
            data = res.json()
            return data
        LOG.error("invalid result for get_attachments status %(status)s",
                  {'status': res.status_code})
        return None

    @http_exc_handler
    def _attach_network(self, fab_parent, nwk_name, lan_attach_list):
        '''
        Function to attach the network.
        '''
        url = self._build_url(self._base_url) + fab_parent + "/" + (
            self._attach_url)
        payload_list = []
        payload = {'networkName' : nwk_name,
                   'lanAttachList': lan_attach_list}
        payload_list.append(payload)
        data = json.dumps(payload_list)
        res = requests.post(url, data=data,
                            headers=self._req_headers,
                            timeout=self._timeout_resp, verify=False)
        LOG.info("attach network url %(url)s res is %(res)s, text is %(text)s, "
                 "reason %(reason)s, data %(data)s",
                 {'url': url, 'res': res, 'text': res.text,
                  'reason': res.reason, 'data': data})
        if res and res.status_code in self._resp_ok:
            LOG.info("attach network successful")
        else:
            LOG.info("attach network failed with status %s, res %s",
                     res.status_code, res.json())

    @http_exc_handler
    def _is_nwk_exist(self, fabric, nwk):
        '''
        Function that returns if the network for the fabric exists
        in DCNM.
        '''
        url = self._build_url(self._nwk_get_url) + fabric + "/networks/" + nwk
        res = requests.get(url, headers=self._req_headers,
                           timeout=self._timeout_resp, verify=False)
        if res and res.status_code in self._resp_ok:
            data = res.json()
            nwk_val = data.get('networkName')
            if nwk_val is None:
                return False
            return nwk_val == nwk
        LOG.error("invalid result for is_nwk_exist for fabric %s nwk %s",
                  fabric, nwk)
        return False

    def is_nwk_exist(self, fabric, nwk):
        '''
        Function that returns if the network for the fabric exists
        in DCNM.
        '''
        try:
            ret = self.login()
            if not ret:
                LOG.error("Failed to login to DCNM")
                return False
            ret = self._is_nwk_exist(fabric, nwk)
            self.logout()
            return ret
        except Exception as exc:
            LOG.error("Exception raised in is_nwk_exist %s", exc)
            return False

    def _compare_exist_cfg_sw(self, exist_cfg_elem, new_cfg, enable):
        '''
        Function to compare the existing attachment config in DCNM
        with the passed config and decide whether to:
        1. Attach/configure => ADD
        2. Don't do anythong => NOOP
        3. unattach followed by attach => DELETE_ADD
        '''
        lan_attach = exist_cfg_elem.get('isLanAttached')
        if enable:
            # Config is not attached, so add everything.
            if not lan_attach:
                return ADD
        else:
            # config is not attached, so nothing to disable.
            if not lan_attach:
                return NOOP
            # No need to perform the remaining checks for disable.
            # ADD refers to proceeding with the operation.
            return ADD

        # 1. When the new config has switches, that is not in the existing
        # config then it can either be a new switch add belonging to an existing
        # network or a change in the neighbor switch itself. This is the case of
        # ADD.
        # 2. When the existing config has switches, that is not in the new
        # config then it is a case of a change in the neighbor switch itself.
        # Should we remove the attachment from the removed switch?? TODO
        #
        # Ideally 1. should be a ADD and 2 should be a DELETE.
        # For now, case 1. is handled in the calling function itself.
        # case 2 is not handled here and a NOOP is returned.
        exist_snum = exist_cfg_elem.get('switchSerialNo')
        new_sw_vlan_intf_info = new_cfg.get('snum_vlan_intf_dict')
        if exist_snum not in new_sw_vlan_intf_info:
            return NOOP
        new_sw_info = new_sw_vlan_intf_info.get(exist_snum)
        # This is the case of VLAN changing, so add the new network. It's not
        # a case of DELETE_ADD.
        # Different networks can still use the same
        # interface with different VLAN's. The below logic doesn't impact
        # that. Only when the same network gets different VLAN's, one needs to
        # Delete and Add.
        if str(exist_cfg_elem.get('vlanId')) != str(new_sw_info.get('vlan')):
            return ADD

        # Now comparing the interface list
        exist_intf = exist_cfg_elem.get('portNames')
        if exist_intf is not None:
            exist_intf_list = exist_intf.split(",")
        else:
            exist_intf_list = []
        new_intf_list = new_sw_info.get('intf_list')
        # 1. If any new interface is not present in the existing config, then
        # it's a case of interface add (may be a new switch add or worst a
        # neighbor change). Interface add for PO will still be a PO, it's not
        # a ADD. We do a ADD.
        # 2. IF any existing interface is not present in the new config, then
        # it's a case of interface change and we need to a DELETE_ADD
        # This may have some duplicate attach for interfaces thar are present
        # in both existing and new configs. This is simpler for now.
        # We will revisit if there's an issue.
        # case 2 first
        if not set(exist_intf_list).issubset(set(new_intf_list)):
            return DELETE_ADD
        # case 1
        if not set(new_intf_list).issubset(set(exist_intf_list)):
            return ADD
        return NOOP

    def _compare_exist_cfg(self, exist_cfg_arr, new_cfg, enable):
        '''
        Function to compare the passed config with existing config in
        DCNM for each switch.
        '''
        sw_oper_dict = {}
        exist_cfg_dict = {}
        if len(exist_cfg_arr) > 0:
            exist_cfg = exist_cfg_arr[0]
            for exist_cfg_elem in exist_cfg.get('lanAttachList'):
                snum = exist_cfg_elem.get('switchSerialNo')
                oper = self._compare_exist_cfg_sw(exist_cfg_elem, new_cfg,
                                                  enable)
                exist_cfg_dict.update({snum: exist_cfg_elem})
                sw_oper_dict.update({snum: oper})
        for snum in new_cfg.get('snum_vlan_intf_dict').keys():
            if snum not in sw_oper_dict:
                if enable:
                    sw_oper_dict.update({snum: ADD})
                else:
                    # For a disable case, if there's a new switch, it's a NOOP
                    sw_oper_dict.update({snum: NOOP})
        return sw_oper_dict, exist_cfg_dict

    def _create_lan_attach_data(self, fabric, nwk_name, snum,
                                sw_ports, vlan, enable):
        dot1q = 1
        lan_attach = {'fabric': fabric, 'networkName': nwk_name,
                      'serialNumber': snum, 'switchPorts': sw_ports,
                      'vlan': vlan, 'dot1QVlan': dot1q, 'untagged': False,
                      'detachSwitchPorts': "", 'freeformConfig': "",
                      'extensionValues': "", 'instanceValues': "",
                      'deployment': enable}
        return lan_attach

    def attach_network(self, enable, deploy, arg_dict):
        '''
        Top level function to attach the network.
        '''
        try:
            ret = self.login()
            if not ret:
                LOG.error("Failed to login to DCNM")
                return
            LOG.info("attach_network arg is %s for enable: %s", arg_dict,
                     enable)
            fab_parent = arg_dict.get('fab_parent')
            fabric = arg_dict.get('fab')
            nwk_name = arg_dict.get('nwk')
            snum_vlan_intf_dict = arg_dict.get('snum_vlan_intf_dict')
            #vlan = arg_dict.get('vlan')

            exist_data = self._get_attachments(fab_parent, nwk_name)
            LOG.info("exist_data is %s", exist_data)
            sw_oper_dict, exist_cfg_dict = self._compare_exist_cfg(exist_data,
                                                                   arg_dict,
                                                                   enable)
            lan_attach_del_list = []
            lan_attach_list = []
            for snum, vlan_intf_dict in snum_vlan_intf_dict.items():
                oper = sw_oper_dict.get(snum)
                LOG.info("For %s, operation is %s", snum, oper)
                if oper == NOOP:
                    continue
                vlan = vlan_intf_dict.get('vlan')
                intf_list = vlan_intf_dict.get('intf_list')
                if enable:
                    if oper == DELETE_ADD:
                        exist_cfg = exist_cfg_dict.get(snum)
                        lan_attach_del = self._create_lan_attach_data(
                            fabric, nwk_name, snum,
                            exist_cfg.get('portNames'),
                            exist_cfg.get('vlanId'), False)
                        lan_attach_del_list.append(lan_attach_del)
                    # For ADD and a DELETE_ADD, the below enable is common
                    lan_attach = self._create_lan_attach_data(
                        fabric, nwk_name, snum,
                        ",".join(intf_list), vlan, enable)
                    lan_attach_list.append(lan_attach)
                else:
                    #disable cannot have a DELETE_ADD
                    lan_attach = self._create_lan_attach_data(
                        fabric, nwk_name, snum,
                        ",".join(intf_list), vlan, enable)
                    lan_attach_list.append(lan_attach)
            flag = False
            if len(lan_attach_del_list) != 0:
                flag = True
                self._attach_network(fab_parent, nwk_name, lan_attach_del_list)
            if len(lan_attach_list) != 0:
                flag = True
                self._attach_network(fab_parent, nwk_name, lan_attach_list)
            if flag and deploy:
                self._deploy_network(fabric, nwk_name)
            self.logout()
        except Exception as exc:
            LOG.error("attach network failed with %(exc)s", {'exc': exc})

    @http_exc_handler
    def _deploy_network(self, fabric, nwk_name):
        '''
        Function to deploy the network in DCNM.
        '''
        url = self._build_url(self._base_url) + fabric + "/" + (
            self._deploy_url)
        payload = {'networkNames' : nwk_name}
        #payload_list.append(payload)
        data = json.dumps(payload)
        res = requests.post(url, data=data,
                            headers=self._req_headers,
                            timeout=self._timeout_resp, verify=False)
        if res and res.status_code in self._resp_ok:
            LOG.info("deploy network successful")
        else:
            LOG.info("deploy network failed with res %s", res)

    def deploy_network(self, fabric, nwk_name):
        '''
        Top level function to deployt the network.
        '''
        try:
            ret = self.login()
            if not ret:
                LOG.error("Failed to login to DCNM")
                return
            self._deploy_network(fabric, nwk_name)
            self.logout()
        except Exception as exc:
            LOG.error("deploy network failed with exception %(exc)s",
                      {'exc': exc})



# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
DCNM Top level class that instantiates the vSphere class based on the config
file. Also, instantiates the DCNM helper classes.
'''

import sys
from workload_auto import logger
from workload_auto import dcnm_helper
from workload_auto import vsph_handler
from workload_auto import utils
from workload_auto import rest

LOG = logger.get_logging(__name__)

class DcnmTop:
    '''
    Top level DCNM class that instantiates other classes.
    '''
    def __init__(self, conf_dict):
        '''
        Init function to instantiate all vsphere classes and calls configure
        in vsphere class that reads the spec file and attaches the networks.
        Finally, it calls the rest initialization routine.
        '''
        LOG.info("Initializing DCNM Top class conf is %s", conf_dict)
        self.conf_dict = conf_dict
        self.dcnm_vsph_dict = {}
        self.auto_deploy = self.conf_dict.get('AutoDeploy')
        if self.auto_deploy is None:
            self.auto_deploy = False
        #self.auto_deploy = auto_deploy_str == "True"
        try:
            ret = self.instantiate_all()
            if not ret:
                return
            # Need to find the right place to put this
            self.init_cfg_all()
            port = self.conf_dict.get('ListenPort')
            rest.rest_init(port, self.rest_handler)
        except Exception as exc:
            LOG.error("Exception raised in DCNM top. Instantiation failed "
                      "with %s", exc)
            return

    def rest_handler(self, op_type, data):
        '''
        Top Level Rest handler to call the appropriate routine based on the
        type.
        '''
        LOG.info("op is %s data %s", op_type, data)
        if op_type == "CLEAN":
            self.cleanup()
        elif op_type == "REFRESH":
            self.refresh()
        elif op_type == "RESYNC":
            self.resync()

    def check_dcnm_cfg(self, nwk_mgr_dict):
        '''
        Function to check if DCNM config has all the mandatory keys.
        '''
        if nwk_mgr_dict.get('Ip') is None or (
                nwk_mgr_dict.get('User') is None) or (
                    nwk_mgr_dict.get('Password') is None) or (
                        nwk_mgr_dict.get('CsvFile') is None):
            return False
        return True

    def check_vsph_cfg(self, vsph_dict):
        '''
        Function to check if vsphere config has all the mandatory keys.
        '''
        if vsph_dict.get('Ip') is None or (
                vsph_dict.get('User') is None) or (
                    vsph_dict.get('Password') is None):
            return False
        return True

    def validate_csv_values(self, csv_list):
        '''
        Validate the contents of CSV file.
        '''
        mand_set = set(['vCenter', 'Fabric', 'Network', 'Dvs', 'Dvs_pg',
                        'Host', 'Host_pg'])
        for csv_dict in csv_list:
            csv_keys = csv_dict.keys()
            if not mand_set.issubset(csv_keys):
                LOG.exception("Mandatory keys %s not present in CSV file",
                              mand_set)
                return False
            vcntr_ip = csv_dict.get('vCenter')
            if not vcntr_ip:
                LOG.exception("vCenter IP is empty in csv file")
                return False
            fabric = csv_dict.get('Fabric')
            if not fabric:
                LOG.exception("Fabric Name is empty in csv file")
                return False
            nwk_name = csv_dict.get('Network')
            if not nwk_name:
                LOG.exception("DCNM Nwk Name is empty in csv file")
                return False
            dvs_name = csv_dict.get('Dvs')
            dvs_pg = csv_dict.get('Dvs_pg')
            host_name = csv_dict.get('Host')
            host_pg = csv_dict.get('Host_pg')
            if not dvs_name and not host_name:
                LOG.exception("Both DVS and Host name is missing")
                return False
            if dvs_name:
                if not dvs_pg:
                    LOG.exception("Invalid DVS PG for DVS %s", dvs_name)
                    return False
            else:
                if not host_pg:
                    LOG.exception("Invalid Host PG for host %s", host_name)
                    return False
        return True

    def validate_cfg(self):
        '''
        Function to validate the config and CSV file. Exit, if error.
        '''
        dcnm_nwk_dict = {}
        for dcnm_ip, dcnm_dict in self.dcnm_vsph_dict.items():
            vsph_dict = dcnm_dict.get('vsph_dict')
            csv_file = dcnm_dict.get('csv_file')
            csv_list, exc = utils.csv_file_read(csv_file)
            if not csv_list:
                LOG.exception("Exception in CSV file read, %s", exc)
                sys.exit(1)
            if not self.validate_csv_values(csv_list):
                LOG.exception("Exception in CSV file contents")
                sys.exit(1)
            for csv_map in csv_list:
                nwk = csv_map.get('Network')
                fabric = csv_map.get('Fabric')
                vsph_dict = dcnm_dict.get('vsph_dict')
                if nwk not in dcnm_nwk_dict.keys():
                    if not dcnm_dict.get('obj').is_nwk_exist(fabric, nwk):
                        LOG.exception("DCNM Network %s unavailable for fab %s",
                                      nwk, fabric)
                        sys.exit(1)
                    dcnm_nwk_dict.update({nwk: True})
            for vsph_ip, vsph_obj in vsph_dict.items():
                if not vsph_obj.validate_cfg(csv_list):
                    LOG.exception("Validate vsphere cfg returned false "
                                  "for %s", vsph_ip)
                    sys.exit(1)
            self.dcnm_vsph_dict[dcnm_ip].update({'csv_list': csv_list})

    def cfg_all(self, read_cfg=False, enable=True):
        '''
        Function that reads the spec file and calls each vsphere object
        to configure the network in DCNM based on the contents of the csv file.
        '''
        if read_cfg:
            self.validate_cfg()
            LOG.info("Config file validation is successful")
        try:
            for dcnm_ip, dcnm_dict in self.dcnm_vsph_dict.items():
                vsph_dict = dcnm_dict.get('vsph_dict')

                for vsph_ip, vsph_obj in vsph_dict.items():
                    LOG.info("Going to configure for %(vsphere)s %(dcnm)s",
                             {'vsphere': vsph_ip, 'dcnm': dcnm_ip})
                    # This is supposed to goto queue
                    vsph_obj.configure(read_cfg=read_cfg, enable=enable,
                                       auto_deploy=self.auto_deploy,
                                       csv_list=dcnm_dict.get('csv_list'))
        except Exception as exc:
            LOG.error("DCNM Top: Exception in cfg all is %s", exc)

    def init_cfg_all(self):
        '''
        Wrapper for the cfg_all function called by the init routine.
        '''
        self.cfg_all(read_cfg=True, enable=True)

    def cleanup(self):
        '''
        Cleanup handler, which will call the routine to unconfig the networks.
        '''
        LOG.info("In cleanup handler")
        self.cfg_all(enable=False)

    def refresh(self):
        '''
        Refresh handler, which will call the routine to re-read the spec file
        and perform the config operation again.
        '''
        LOG.info("In refresh handler")
        self.cfg_all(read_cfg=True, enable=True)

    def resync(self):
        '''
        Resync handler, which will call the routine to perform the config
        operation again if certain parameters like the neighbour's networks
        have changed.
        '''
        LOG.info("In resync handler")
        self.cfg_all(enable=True)

    def instantiate_all(self):
        '''
        Function that instantiates all the vsphere objects and DCNM helper
        objects.
        '''
        nwk_mgr_list = self.conf_dict.get('NwkMgr')
        flag = False
        if nwk_mgr_list is None:
            LOG.error("Incorrect yaml format, NwkMgr key not present")
            return flag
        for nwk_mgr_dict in nwk_mgr_list:
            LOG.info("In Instantiate, nwk_mgr_dict is %s", nwk_mgr_dict)
            ret = self.check_dcnm_cfg(nwk_mgr_dict)
            if not ret:
                LOG.error("Invalid Ip or User or Pwd in yaml file, dict")
                continue
            dcnm_ip = nwk_mgr_dict.get("Ip")
            LOG.info("DCNM IP is %s", dcnm_ip)
            dcnm_obj = dcnm_helper.DcnmHelper(ip=dcnm_ip,
                                              user=nwk_mgr_dict.get("User"),
                                              pwd=nwk_mgr_dict.get("Password"))
            csv_file = nwk_mgr_dict.get("CsvFile")
            vsph_list = nwk_mgr_dict.get('ServerCntrlr')
            if vsph_list is None:
                LOG.error("Incorrect yaml, ServerCntrlr key not present")
                continue
            self.dcnm_vsph_dict.update({dcnm_ip: {'obj': dcnm_obj,
                                                  'csv_file': csv_file,
                                                  'vsph_dict': {}}})
            for vsph_dict in vsph_list:
                ret = self.check_vsph_cfg(vsph_dict)
                if not ret:
                    LOG.error("Invalid values for server ctrlr in yaml file %s",
                              vsph_dict)
                    continue
                vsph_ip = vsph_dict.get("Ip")
                vsph_obj = vsph_handler.VsphHandler(vsph_dict=vsph_dict,
                                                    dcnm_obj=dcnm_obj)
                self.dcnm_vsph_dict[dcnm_ip]['vsph_dict'].update(
                    {vsph_ip: vsph_obj})
                flag = True
        LOG.info("Instantiated successfully, dict is %s", self.dcnm_vsph_dict)
        return flag



# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
Simple logger module.
'''

import logging
import logging.handlers as log_hdlr

LOG_FRMT = '%(asctime)s %(levelname)8s [%(name)s] %(message)s'
LOG_DATE_FRMT = '%Y-%m-%d %H:%M:%S'
# 5MB
MAX_LOG_SIZE = 5000000
LOG_BACKUP_CNT = 10

_logger_dict = {}

def get_logging(name):
    '''
    Function to get the logger object.
    '''
    if name in _logger_dict:
        return _logger_dict[name]
    logger = logging.getLogger(name)
    _logger_dict[name] = logger
    return logger

def set_logging(filename, level=None):
    '''
    Main set logger function.
    '''
    #logging.basicConfig(filename=filename)
    #, level=level, format='%(asctime)s %(message)s')
    if level is None:
        level = logging.INFO
    log_format = LOG_FRMT
    log_date_format = LOG_DATE_FRMT
    log_prnt_format = logging.Formatter(fmt=log_format, datefmt=log_date_format)
    logger = logging.getLogger(None)
    logger.setLevel(level)
    handler = log_hdlr.RotatingFileHandler(filename, maxBytes=MAX_LOG_SIZE,
                                           backupCount=LOG_BACKUP_CNT)
    handler.setFormatter(log_prnt_format)
    logger.addHandler(handler)



# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
vSphere/vCenter helper module. This module interacts with vSphere using pynmoni
'''

from pyVim.connect import SmartConnectNoSSL
from pyVmomi import vim
from workload_auto import logger

LOG = logger.get_logging(__name__)

class VsphereHelper:
    '''
    VSphere Helper class
    '''
    def __init__(self, **kwargs):
        '''
        Init routine that connects with the specified vCenter
        '''
        self.ip_addr = kwargs.get('ip')
        self.user = kwargs.get('user')
        self.pwd = kwargs.get('pwd')
        self.si_obj = SmartConnectNoSSL(host=self.ip_addr, user=self.user,
                                        pwd=self.pwd)

    def get_all_objs(self, vimtype):
        '''
        Retrieve all the objects of a specific type
        '''
        obj = {}
        container = self.si_obj.content.viewManager.CreateContainerView(
            self.si_obj.content.rootFolder, vimtype, True)
        for managed_object_ref in container.view:
            obj.update({managed_object_ref: managed_object_ref.name})
        return obj

    def get_specific_obj(self, vimtype, name):
        '''
        Retrieve the object of a specific type matching the name.
        '''
        container = self.si_obj.content.viewManager.CreateContainerView(
            self.si_obj.content.rootFolder, vimtype, True)
        for managed_object_ref in container.view:
            if managed_object_ref.name == name:
                return managed_object_ref
        return None

    def get_dvs_pg_obj(self, dvs_name, dvs_pg):
        '''
        Returns the DVS and DVS PG objects.
        '''
        dvs_obj = self.get_specific_obj([vim.DistributedVirtualSwitch],
                                        dvs_name)
        if dvs_obj is None or dvs_obj.name != dvs_name:
            LOG.error("DVS %s does not exist", dvs_name)
            return None, None
        dvs_pgs = dvs_obj.portgroup
        for dvpg in dvs_pgs:
            if dvpg.name == dvs_pg:
                return dvs_obj, dvpg
        LOG.error("DVS PG %s does not exist", dvs_pg)
        return None, None

    def is_dvs_dvspg_exist(self, dvs_name, dvs_pg):
        '''
        Checks if the DVS and DVS PG exist in vSphere.
        '''
        dvs_obj, dvs_pg_obj = self.get_dvs_pg_obj(dvs_name, dvs_pg)
        if dvs_obj is None or dvs_pg_obj is None:
            LOG.error("get_dvs_pg_obj returns false for DVS %s dvs PG %s",
                      dvs_name, dvs_pg)
            return False
        return True

    def get_host_pg_obj(self, host_name, host_pg):
        '''
        Returns the Host and PG object
        '''
        host_obj = self.get_specific_obj([vim.HostSystem], host_name)
        if host_obj is None or host_obj.name != host_name:
            LOG.error("Host %s does not exist", host_name)
            return None, None
        for pg_obj in host_obj.config.network.portgroup:
            if pg_obj.spec.name == host_pg:
                return host_obj, pg_obj
        LOG.error("Host PG %s does not exist", host_pg)
        return None, None

    def is_host_hostpg_exist(self, host_name, host_pg):
        '''
        Checks if the Host and Host PG exist in vSphere.
        '''
        host_obj, host_pg_obj = self.get_host_pg_obj(host_name, host_pg)
        if host_obj is None or host_pg_obj is None:
            LOG.error("get_host_pg_obj returns false for host %s host PG %s",
                      host_name, host_pg)
            return False
        return True

    def get_vlan_dvs(self, dvs_name, dvs_pg):
        '''
        Get the Vlan associated wih the DV-PG in a DVS.
        First get all the DVS and get the DVS object matching the argument DVS
        For all the DV-PG in the DVS object, find the DV-PG that matches the
        DV-PG needed
        Then from that DV-PG object get the VLAN as give.
        Wish there's a should be a faster method to directly get the dvpg
        info instead of going through the loop.
        '''
        dvs_obj, dvs_pg_obj = self.get_dvs_pg_obj(dvs_name, dvs_pg)
        if dvs_obj is None or dvs_pg_obj is None:
            LOG.error("DVS %s or PG %s does not exist, cannot obtain VLAN",
                      dvs_name, dvs_pg)
            return ""
        vlan_info = dvs_pg_obj.config.defaultPortConfig.vlan
        cl_obj = vim.dvs.VmwareDistributedVirtualSwitch.TrunkVlanSpec
        if isinstance(vlan_info, cl_obj):
            # This needs to be tested
            return ""
        vlan_id = str(vlan_info.vlanId)
        return vlan_id

    def get_vlan_host_pg(self, host_name, pg_name):
        '''
        Get the Vlan associated wih the PG in a Host.
        First get the network object associated with the PG.
        Second, match the host, in case the same PG name is cfgs in multiple
        hosts. There are other ways to write the loop, but i think this is
        little better since the host loop being big is rare unless one cfgs the
        same PG name in multiple hosts.
        Next, for all the PG cfgd in the host find the matching PG object and
        return the VLAN.
        '''
        #hosts = self.get_all_objs([vim.HostSystem])
        host_obj, host_pg_obj = self.get_host_pg_obj(host_name, pg_name)
        if host_obj is None or host_pg_obj is None:
            LOG.error("Host %s or PG %s does not exist, cannot obtain VLAN",
                      host_name, pg_name)
            return ""
        vlan_id = str(host_pg_obj.spec.vlanId)
        return vlan_id

    def get_vlan(self, is_dvs, dvs_or_host_name, cmn_pg):
        '''
        Top level get vlan function.
        '''
        if is_dvs:
            return self.get_vlan_dvs(dvs_or_host_name, cmn_pg)
        else:
            return self.get_vlan_host_pg(dvs_or_host_name, cmn_pg)

    def get_host_neighbour(self, host_name, pnic_list):
        '''
        Retrieve the neighbour given a host and pnic's.
        '''
        host = self.get_specific_obj([vim.HostSystem], host_name)
        neighbour = []
        query = host.configManager.networkSystem.QueryNetworkHint()
        for nic in query:
            if not hasattr(nic, 'device'):
                LOG.error('device attribute not present')
                continue
            pnic = nic.device
            if pnic not in pnic_list:
                continue
            dct = {}
            conn_sw_port = nic.connectedSwitchPort
            if not hasattr(conn_sw_port, 'portId') or (
                    not hasattr(conn_sw_port, 'devId')):
                LOG.error("Port Id or devId attribute not present")
                continue
            sw_port = conn_sw_port.portId
            dev_id = conn_sw_port.devId
            snum_str = dev_id.split('(')
            if len(snum_str) > 1:
                snum = snum_str[1].split(')')[0]
            else:
                LOG.error("snum not present for the switch")
            if hasattr(conn_sw_port, 'mgmtAddr'):
                sw_ip = conn_sw_port.mgmtAddr
            if hasattr(conn_sw_port, 'systemName'):
                sw_name = conn_sw_port.systemName
            dct.update({'ip': sw_ip, 'snum': snum, 'pnic': pnic,
                        'sw_port': sw_port, 'name': sw_name})
            neighbour.append(dct)
        return neighbour

    def get_pnic_dvs(self, host_name, key_dvs):
        '''
        Return the list of pnic's of a host that is a part of a DVS.
        The pnic is taken from the host of DCV object.
        '''
        pnic_list = []
        for host in key_dvs.config.host:
            if host.config.host.name == host_name:
                for pnic_elem in host.config.backing.pnicSpec:
                    pnic_list.append(pnic_elem.pnicDevice)
        return pnic_list

    def get_dvs_pnic_info(self, dvs, dvs_pg):
        '''
        Return the list of neighbours for every host associated with the DVS.
        First get the matching DVS object and DV-PG object from the complete
        list.
        For the host that is a part of the DVS get the pnic and neighbour info.
        '''
        dvs_obj, dvs_pg_obj = self.get_dvs_pg_obj(dvs, dvs_pg)
        if dvs_obj is None or dvs_pg_obj is None:
            LOG.error("DVS %s or PG %s does not exist, cannot obtain pnic",
                      dvs, dvs_pg)
            return ""
        host_dict = {}
        for host_obj in dvs_pg_obj.host:
            pnic_list = self.get_pnic_dvs(host_obj.name, dvs_obj)
            nei_list = self.get_host_neighbour(host_obj.name, pnic_list)
            host_dict.update({host_obj.name: nei_list})
        return host_dict

    def _get_pnic_from_key(self, pnic_obj, pnic_comp_key):
        '''
        Return the device from pnic object for matching object name.
        '''
        for pnic_elem in pnic_obj:
            if pnic_elem.key == pnic_comp_key:
                return pnic_elem.device
        return None

    def get_host_pnic_info(self, host_name, pg_name):
        '''
        Get the pnic info for a specific host.
        First get the list of all network object for the PG.
        Then, filter the objects based on the passed host. This, i assume
        may not be much unless the same PG name is cfgd in multiple host.
        Get the vSwitch name that this PG is a part of.
        Then, for all the vswitches in the host, filter the specific vswitch
        and get the vswitch object.
        Get the pnic list from the vswitch, this gives the object list!!
        Then call _get_pnic_from_key which returns the dev associated
        with the pnic object.
        Retrieve the neighbour for the host and pnic.
        '''
        #hosts = self.get_all_objs([vim.HostSystem])
        host_dict = {}
        host_obj, host_pg_obj = self.get_host_pg_obj(host_name, pg_name)
        if host_obj is None or host_pg_obj is None:
            LOG.error("Host %s or PG %s does not exist, cannot obtain pnic",
                      host_name, pg_name)
            return ""
        vsw_name = host_pg_obj.spec.vswitchName
        for vsw in host_obj.config.network.vswitch:
            if vsw.name != vsw_name:
                continue
            pnic_list = vsw.pnic
            pnic_dev_list = []
            for pnic_elem in pnic_list:
                dev = self._get_pnic_from_key(host_obj.config.network.pnic,
                                              pnic_elem)
                pnic_dev_list.append(dev)
            nei_list = self.get_host_neighbour(host_name, pnic_dev_list)
            host_dict.update({host_name: nei_list})
        return host_dict

    def get_pnic(self, is_dvs, dvs_host_name, cmn_pg):
        '''
        Top level function to retrieve the information associated with the pnic.
        '''
        if is_dvs:
            return self.get_dvs_pnic_info(dvs_host_name, cmn_pg)
        else:
            return self.get_host_pnic_info(dvs_host_name, cmn_pg)



# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''
Setup file.
'''
#from distutils.core import setup
import os
from setuptools import setup, find_packages

def read(fname):
    '''
    Read the file contents for lomng description.
    '''
    with open(os.path.join(os.path.dirname(__file__), fname),
              encoding='utf-8') as file_desc:
        file_data = file_desc.read()
    str_var = ""
    for file_ln in file_data.split("\n"):
        if not file_ln.startswith("#"):
            str_var = str_var + file_ln + "\n"
    return str_var

vmm_env = os.getenv("VMM_TMP_VENV")
if vmm_env is None:
    CFG_DIR = '/etc/vmm_workload_auto'
else:
    CFG_DIR = './etc/vmm_workload_auto'

setup(name='vmm_workload_auto',
      version='1.0.2',
      description='Workload Automation for VMM',
      author='Padmanabhan Krishnan',
      author_email='padkrish@cisco.com',
      url='https://www.cisco.com',
      long_description=read('./README.txt'),
      long_description_content_type='text/x-rst',
      #package_dir={'vmm_workload_auto': 'workload_auto'},
      #packages=['vmm_workload_auto'],
      packages=find_packages(exclude=['tests']),
      include_package_data=True,
      #data_files=[('/etc/vmm_workload_auto',
      data_files=[(CFG_DIR,
                   ['config/conf.yml',
                    'config/sample.csv',
                    'config/conf_multiple_dcnm.yml',
                    'config/conf_multiple_vcenter.yml'])],
      install_requires=[
          'pyvim', 'pyvmomi', 'PyYAML', 'Flask',
      ],
      entry_points={
          'console_scripts': ['vmm_workload_auto = workload_auto.main:wl_auto_main']
      },
      scripts=['setup.sh'],
      zip_safe=False,)



# Copyright 2020 Cisco Systems, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

'''

File containing general utility functions.
'''
import csv
import yaml
from workload_auto import logger

LOG = logger.get_logging(__name__)

def csv_file_read(filename):
    '''
    Function to read the CSV file.
    '''
    try:
        with open(filename, newline='') as csvfile:
            csvfile.seek(0)
            reader = csv.DictReader(csvfile)
            csv_list = []
            for row in reader:
                csv_map = {}
                for key, val in row.items():
                    csv_map.update({key.strip(): val.strip()})
                csv_list.append(csv_map)
            return csv_list, None
    except Exception as exc:
        LOG.error("CSV file read exception %s", exc)
        return [], exc

def yml_file_read(filename):
    '''
    Function to read the YML file.
    '''
    try:
        with open(filename, 'r') as file_desc:
            file_dict = yaml.full_load(file_desc)
            return file_dict, None
    except Exception as exc:
        LOG.error("YML file read exception %s", exc)
        return {}, exc
